/* robin.c -- round-robin iteration to improve the score of a multiple
*  alignment.  The entry point is:
*	align_ptr robin(align_ptr A, int nRow, piece_ptr P, int nP, int width,
*	  ss_t ss, int *gop, int *gtype, int max_iter, int max_depth)
*  where A is the alignment, in which columns stored in a linked list.
*/

#include "refine.h"

#define CAUTIOUS

typedef struct interv {
	align_ptr left, right;
	int depth;
	struct interv *next;
} *interval_ptr, interval;

static void pr_align(align_ptr from, align_ptr to, int nRow)
{
	align_ptr ap;
	int k;

	for (ap = from; ap != to->next; ap = ap->next)
		fputc((ap->unchanged ? '+' : '-'), stderr);
	fputc('\n', stderr);
	for (k = 0; k < nRow; ++k) {
		for (ap = from; ap != to->next; ap = ap->next)
			fprintf(stderr, "%c", ap->col[k]);
		fputc('\n', stderr);
	}
}

int order_align(align_ptr ap)
{
	int i;

	for (i = 0; ap; ap = ap->next)
		ap->colno = ++i;
	return i;
}

/* subalign -- Create a 2-dimensional array from the linked list of columns
*  from "left" to "right", omitting the designated row; set column bounds LB
*  and RB; create the "Gap-Column" array GC, which gives DASH/FREE columns
*  that can be interposed between columns of the packed alignment; extract
*  the sequence to be aligned to the 2-D array.
*/
static uchar **subalign(align_ptr left, align_ptr right, int nRow,
  int pi, int width, piece_ptr P, int nP, int *LB, int *RB,
  uchar ***GCptr, int *nCol_ptr, uchar **Seq_ptr, int *seqlen_ptr)
{
	uchar **AL, **GC, *acol, *ncol, *Seq;
	align_ptr ap;
	int i, j, k, p, r, seqpos, nCol, lo, hi, x, y, row = P[pi].row;

	nCol = right->colno - left->colno + 1;	/* some may not be needed */
	--nRow;

	/* Allocate memory for AL, GC and Seq */
	AL = (uchar **)ckalloc(nCol*sizeof(uchar *)) - 1;
	AL[1] = (uchar *)ckalloc(nRow*nCol*sizeof(uchar));
	for (i = 2; i <= nCol; ++i)
		AL[i] = AL[i-1] + nRow;
	GC = (uchar **)ckalloc((nCol+1)*sizeof(uchar *));
	GC[0] = (uchar *)ckalloc(nRow*(nCol+1)*sizeof(uchar));
	for (i = 1; i <= nCol; ++i)
		GC[i] = GC[i-1] + nRow;
	Seq = (uchar *)ckalloc(nCol*sizeof(uchar)) - 1;

	/* Scan from "left" to "right"; set AL, Seq, LB and RB; set "subcol"
	*  mapping from linked columns in A to columns of AL, which is used both
	*  here and in replace_cols().  */
	seqpos = 0;
	LB[0] = 0;
	RB[0] = width;
	for (i = 0, ap = left; ap && ap->colno <= right->colno; ap = ap->next) {
		acol = ap->col;
		if (isalpha(acol[row]))
			Seq[++seqpos] = acol[row];
		for (k = 0; k <= nRow; ++k)
			if (k != row && isalpha(acol[k]))
				break;
		if (k > nRow) {	/* .. only gaps outside of "row" */
			RB[i]++;
			ap->subcol = -1;
		} else {
			++i;
			ap->subcol = i;
			if (i > nCol)
				fatal("error 1 in subalign");
			ncol = AL[i];
			LB[i] = RB[i] = seqpos;
			for (j = k = 0; k <= nRow; ++k)
				if (k != row)
					ncol[j++] = acol[k];
		}
	}
	nCol = i;
	if (width < 0)
		fatal("width must be non-negative");
	width = MIN(width, nCol);
	for (i = 0; i < nCol - width; ++i)
		RB[i] = MIN(seqpos, RB[i+width] + width);
	for (i = nCol - width; i <= nCol; ++i)
		RB[i] = seqpos;
	for (i = nCol; i > width; --i)
		LB[i] = MAX(0, LB[i-width] - width);
	for (i = 0; i <= width; ++i)
		LB[i] = 0;

	/* Initialize GC to all FREE; for each "piece", replace FREE from
	*  firstcol to lastcol in the appropriate row by DASH. */
	for (i = 0; i <= nCol; ++i)
		for (k = 0; k < nRow; ++k)
			GC[i][k] = FREE;
	lo = left->colno;
	hi = right->colno;
	for (p = 0; p < nP; ++p) {
		if (p == pi ||
		    P[p].firstcol->colno > hi || P[p].lastcol->colno < lo)
			continue;
		r = P[p].row;
		if (r > row)
			--r;
		x = (P[p].firstcol->colno < lo ? 0 : P[p].firstcol->subcol);
		y = (P[p].lastcol->colno > hi ? nCol+1 : P[p].lastcol->subcol);
		for (i = x; i < y; ++i)
			GC[i][r] = DASH;
	}

	/* Return */
	*GCptr = GC;
	*nCol_ptr = nCol;
	*Seq_ptr = Seq;
	*seqlen_ptr = seqpos;
	return AL;
}

static int column_type(uchar *col, int row, int nRow)
{
	int r;

	for (r = 0; r < nRow; ++r)
		if (r != row && isalpha(col[r]))
			break;
	if (isalpha(col[row]))
		return (r < nRow ? MATCH_COL : GAP_AL);
	else if (r < nRow)
		return GAP_SEQ;
	else
		fatal("column_type() impossible");
}

/* replace_cols -- Replace linked columns from "left" to "right" by columns
*  from the 2-D array AL; update P. */
static void replace_cols(align_ptr *A, align_ptr left, align_ptr right,
  uchar **AL, int nRow, int nCol, int row, int pi, piece_ptr P, int nP,
  align_ptr *Left_ptr, align_ptr *Right_ptr)
{
	int i, r, lo, hi, p, n, alpos1, alpos2, seqpos1, seqpos2;
	uchar *c, *ncol, *acol;
	align_ptr ap, bp, *newcol;

	/* The values align_ptr->subcol map linked columns between left and
	*  right into column indices in the packed alignment created by
	*  subalign().  We first replace those values by column indices in
	*  the new alignment, AL.
	*/
	for (i = 0, ap = left; ap != right->next; ap = ap->next) {
		if (ap == NULL)
			fatal("replace_cols() -- no way");
		if (ap->subcol < 0)
			continue;
		for (r = 0; r < nRow; ++r)
			if (r != row && isalpha(ap->col[r]))
				break;
		if (r == nRow)
			fatal("Cannot translate alignment row indices.");
		while (++i <= nCol && !isalpha(AL[i][r]))
			;
		if (i > nCol)
			fatal("replace_cols bad 1.");
		if (AL[i][r] != ap->col[r])
			fatal("replace_cols bad 2.");
		ap->subcol = i;
	}

#ifdef CAUTIOUS
	while (++i <= nCol)
		if (column_type(AL[i], row, nRow) != GAP_AL)
			fatal("Problem encountered in subalign.");
#endif

	/* Create the new piece of alignment.*/
	newcol = (align_ptr *)ckalloc(nCol*sizeof(align_ptr)) - 1;
	for (i = 1; i <= nCol; ++i) {
		newcol[i] = (align_ptr)ckalloc(sizeof(*ap));
		c = newcol[i]->col = (uchar *)ckalloc(nRow*sizeof(uchar));
		for (r = 0; r < nRow; ++r)
			c[r] = AL[i][r];
		if (i > 1) {
			newcol[i]->prev = newcol[i-1];
			newcol[i-1]->next = newcol[i];
		}
	}
	/* Splice the new piece into the large alignment. */
	newcol[1]->prev = left->prev;
	if (left == *A)
		*A = newcol[1];
	else
		left->prev->next = newcol[1];
	newcol[nCol]->next = right->next;
	if (right->next)
		right->next->prev = newcol[nCol];
	else
		(*A)->prev = newcol[nCol];

	/* Update firstcol and lastcol whenever the end of a piece falls at
	*  a replaced column. */
	lo = left->colno;
	hi = right->colno;
	for (p = 0; p < nP; ++p) {
		if (p == pi)
			continue;
		n = P[p].firstcol->colno;
		if (lo <= n && n <= hi)
			P[p].firstcol = newcol[P[p].firstcol->subcol];
		n = P[p].lastcol->colno;
		if (lo <= n && n <= hi)
			P[p].lastcol = newcol[P[p].lastcol->subcol];
	}

	/* Determine firstcol and lastcol for the row that was adjusted. */
	if (P[pi].firstcol->colno >= lo) {
		for (i = 1; i <= nCol; ++i)
			if (isalpha(newcol[i]->col[row]))
				break;
#ifdef CAUTIOUS
		if (i > nCol)
			fatal("replace_cols: cannot happen 1");
#endif
		P[pi].firstcol = newcol[i];
	}
	if (P[pi].lastcol->colno <= hi) {
		for (i = nCol; i >= 1; --i)
			if (isalpha(newcol[i]->col[row]))
				break;
#ifdef CAUTIOUS
		if (i < 1)
			fatal("replace_cols: cannot happen 2");
#endif
		P[pi].lastcol = newcol[i];
	}
	
	/* Determine which columns are unchanged.  Scan the new columns and the
	*  old columns in parallel, keeping track of positions in the extracted
	*  subalignment and the extracted Seq that have just now been aligned.*/
	ap = NULL;
	acol = NULL;	/* not used */
	alpos1 = alpos2 = seqpos1 = seqpos2 = 0;
	for (i = 1; i <= nCol; ++i) {
		/* updata alpos1 and seq1pos for the new column */
		ncol = newcol[i]->col;
		for (r = 0; r < nRow; ++r)
			if (r != row && isalpha(ncol[r]))
				break;
		if (r < nRow)
			++alpos1;
		if (isalpha(ncol[row]))
			++seqpos1;
		/* Find old column with similar positions. */
		while (alpos2 < alpos1 || seqpos2 < seqpos1) {
			ap = (ap == NULL ? left : ap->next);
			acol = ap->col;
			if (ap->subcol > 0)
				++alpos2;	/* initial ap->subcol? */
			if (isalpha(acol[row]))
				++seqpos2;
		}

		newcol[i]->unchanged =
		   (alpos1 == alpos2 && seqpos1 == seqpos2 &&
		    column_type(ncol,row,nRow) == column_type(acol,row,nRow)) ?
			ap->unchanged : 0;

#ifdef CAUTIOUS
		if (newcol[i]->unchanged) 
			for (r = 0; r < nRow; ++r)
				if (ncol[r] != acol[r])
					fatal("Is this impossible?");
#endif
	}
			

	/* Free the spliced-out columns. */
	for (ap = left; ; ap = bp) {
		free(ap->col);
		bp = ap->next;
		free(ap);
		if (ap == right)
			break;
	}
	
	*Left_ptr = newcol[1];
	*Right_ptr = newcol[nCol];
	free(newcol+1);
}

static interval_ptr new_interval(align_ptr lf, align_ptr rt, interval_ptr ip,
int depth)
{
	interval_ptr new = (interval_ptr) ckalloc(sizeof(interval));

#ifdef DEBUG
fprintf(stderr, "new interval, columns %d-%d\n", lf->colno, rt->colno);
#endif
	new->left = lf;
	new->right = rt;
	new->next = ip;
	new->depth = depth;
	return new;
}

#define RUN_SAME 100	/* minimum number of identical columns to be frozen */
#define RETREAT 10	/* distance of endpoint in run of identical columns */

/* find_changed -- put changed subintervals on the stack */
static interval_ptr find_changed(align_ptr left, align_ptr right,
  interval_ptr I, int depth)
{
	align_ptr a, b, c, rt, lf;
	int ident_cols, i;

#ifdef DEBUG
fprintf(stderr, "find_changed, cols %d-%d\n", left->colno, right->colno);
for (a = left; a != right->next; a = a->next)
fputc(a->unchanged ? '+' : '-', stderr);
fputc('\n', stderr);
#endif
	for (a = right; ; a = a->prev) {
		/* skip long runs of identical columns to find a right end */
		for (b = a; b != left; b = b->prev)
			if (!(b->unchanged))
				break;
		if (b->unchanged)
			break;
		if (b->colno + RUN_SAME <= a->colno)
			for (rt = b, i = 0; i < RETREAT; ++i)
				rt = rt->next;
		else
			rt = a;
		/* search for another such long run, defining the left end */
		for (c = a = b, ident_cols = 0; a != left; a = a->prev) 
			if (a->unchanged) {
				if (++ident_cols >= RUN_SAME)
					break;
			} else {
				ident_cols = 0;
				c = a;
			}
		
		if (ident_cols >= RUN_SAME)
			for (lf = c, i = 0; i < RETREAT; ++i)
				lf = lf->prev;
		else
			lf = left;
		I = new_interval(lf, rt, I, depth+1);
		if (a == left)
			break;
	}
	return I;
}

/* round robin realigner -- 
*    Iteratively remove one piece (of sequence) and realign the sequence with
*    the subalignment consisting of the remaining sequences.
*/
align_ptr robin(align_ptr IA, int nRow, piece_ptr P, int nP, int width,
  ss_t ss, int *gop, int *gtype, int max_iter, int max_depth, TBack tback)
{
	int p, LB_size, curr_score, new_score, test_score, row,
		nCol, newCol, *LB, *RB, iter, lo, hi, seqlen, deepest = 0,
		colno, depth, left_type = 0, right_type = 0;
#ifdef CAUTIOUS
	int correct = 0, incorrect = 0;
#endif
	uchar **AL, **GC, **AL_new, *Seq;
	align_ptr a, left = NULL, right = NULL, A = IA, Left, Right, new_left,
		new_right;
	interval_ptr Interv, temp_int;

	Interv = new_interval(A, A->prev, (interval_ptr) NULL, 0);

	nCol = order_align(A);
	LB_size = sizeof(int)*2*nCol;
	LB = (int *)ckalloc(LB_size);
	RB = (int *)ckalloc(LB_size);
	for (iter = 0; Interv != NULL && iter < max_iter; ++iter) {
		Left = Interv->left;
		Right = Interv->right;
		depth = Interv->depth;
		temp_int = Interv;
		Interv = Interv->next;
		free(temp_int);
		if (depth > max_depth)
			continue;
		deepest = MAX(depth, deepest);
		for (a = Left; a != Right->next; a = a->next)
			a->unchanged = 1;
		lo = Left->colno;
		hi = Right->colno;

		for (p = nP - 1; p > 0; --p) {
			if (P[p].lastcol->colno < lo ||
			    P[p].firstcol->colno > hi)
				continue;
			row = P[p].row;
			if (P[p].firstcol->colno < lo) {
				left = Left;
				left_type = column_type(Left->col, row, nRow);
			} else {
				left = P[p].firstcol;
				colno = MAX(lo, left->colno - 4*width);
				for ( ; left->colno > colno; left = left->prev)
				      if (left->prev->col[row] != FREE)
					break;
				left_type = column_type(left->col, row, nRow);
				if (left_type != GAP_AL)
					left_type = FREE_END;
			}
			if (P[p].lastcol->colno > hi) {
				right = Right;
				right_type = column_type(Right->col, row, nRow);
			} else {
				right = P[p].lastcol;
				colno = MIN(hi, right->colno + 4*width);
				for ( ; right->colno < colno;
				    right = right->next)
					if (right->next->col[row] != FREE)
					  break;
				right_type = column_type(right->col, row, nRow);
				if (right_type != GAP_AL)
				right_type = FREE_END;
			}
			curr_score = align_score(left, right, nRow, ss, gop,
			  gtype);
			AL = subalign(left, right, nRow, p, width, P, nP,
			   LB, RB, &GC, &newCol, &Seq, &seqlen);
			if (newCol >= LB_size)
				fatal("Needs more space for LB and RB.");

			new_score = row_yama(AL, GC, nRow-1, newCol, Seq,
			  seqlen, left_type, right_type, LB, RB, ss, gop,
			  gtype, row, tback, &AL_new, &newCol);

#ifdef CAUTIOUS
			test_score = row_score(AL_new, nRow, newCol, row, ss,
			  gop, gtype);
			if (new_score != test_score) {
				fprintf(stderr,
				  "col %d (type %d) to col %d (type %d)\n",
				  left->colno, left_type,
				  right->colno, right_type);
				fprintf(stderr,
				  "  row = %d, yama = %d, real score = %d\n",
				   row, new_score, test_score);
				++incorrect;
			} else
				++correct;
#endif
			test_score = packed_score(AL_new, nRow, newCol, ss,
			  gop, gtype);
			if (test_score > curr_score) {
				replace_cols(&A, left, right, AL_new, nRow,
				  newCol, row, p, P, nP, &new_left, &new_right);
				nCol = order_align(A);
				if (left == Left)
					Left = new_left;
				if (right == Right)
					Right = new_right;
				lo = Left->colno;
				hi = Right->colno;
			}
			free(AL[1]);
			free(AL+1);
			free(GC[0]);
			free(GC);
			free(Seq+1);
			free(AL_new[1]);
			free(AL_new+1);
		}
		Interv = find_changed(Left, Right, Interv, depth);
	}
	free(LB);
	free(RB);

#ifdef CAUTIOUS
	if (incorrect)
		fprintf(stderr, "%d correct, %d incorrect\n",
		   correct, incorrect);
#endif
//	fprintf(stderr, "%d iterations, deepest %d\n", iter, deepest);

	return A;
}
